package gov.anl.aps.dm.portal.controllers;

import gov.anl.aps.dm.portal.exceptions.DmPortalException;
import gov.anl.aps.dm.portal.model.beans.AbstractFacade;
import gov.anl.aps.dm.portal.model.entities.CloneableEntity;
import gov.anl.aps.dm.common.utilities.CollectionUtility;
import gov.anl.aps.dm.portal.utilities.SessionUtility;

import java.io.Serializable;

import java.util.List;
import javax.annotation.PostConstruct;
import javax.faces.model.DataModel;
import javax.faces.model.ListDataModel;
import javax.faces.model.SelectItem;
import org.apache.log4j.Logger;
import org.primefaces.component.datatable.DataTable;

public abstract class CrudEntityController<EntityType extends CloneableEntity, FacadeType extends AbstractFacade<EntityType>> implements Serializable
{

    private static final Logger logger = Logger.getLogger(CrudEntityController.class.getName());


    protected EntityType current = null;

    private DataModel listDataModel = null;
    private DataTable listDataTable = null;
    private List<EntityType> filteredObjectList = null;


    private DataModel selectDataModel = null;
    private DataTable selectDataTable = null;
    private List<EntityType> selectedObjectList = null;

    
    public CrudEntityController() {
    }

    @PostConstruct
    public void initialize() {
    }

    protected abstract FacadeType getFacade();

    protected abstract EntityType createEntityInstance();

    public abstract String getEntityTypeName();

    public String getDisplayEntityTypeName() {
        return getEntityTypeName();
    }

    public abstract String getCurrentEntityInstanceName();

    public EntityType getCurrent() {
        return current;
    }

    public void setCurrent(EntityType current) {
        this.current = current;
    }

    public void selectByRequestParams() {
    }

    public EntityType getSelected() {
        if (current == null) {
            current = createEntityInstance();
        }
        return current;
    }

    public String resetList() {
        logger.debug("Resetting list");
        return prepareList();
    }

    public String prepareList() {
        logger.debug("Preparing list");
        resetListDataModel();
        return "list?faces-redirect=true";
    }


    public DataTable getListDataTable() {
        if (listDataTable == null) {
            logger.debug("Recreating data table");
            listDataTable = new DataTable();
        }
        return listDataTable;
    }

    public void setListDataTable(DataTable listDataTable) {
        this.listDataTable = listDataTable;
    }

    public DataTable getSelectDataTable() {
        return selectDataTable;
    }

    public void setSelectDataTable(DataTable selectDataTable) {
        this.selectDataTable = selectDataTable;
    }

    public String prepareView(EntityType entity) {
        clear();
        logger.debug("Preparing view");
        current = entity;
        return "view?faces-redirect=true";
    }
    
    public void clear() {
        
    }

    public String view() {
        return "view?faces-redirect=true";
    }

    public String prepareCreate() {
        current = createEntityInstance();
        return "create?faces-redirect=true";
    }
    
    protected void prepareEntityInsert(EntityType entity) throws DmPortalException {
    }

    public String create() {
        try {
            EntityType newEntity = current;
            prepareEntityInsert(current);
            getFacade().create(current);
            SessionUtility.addInfoMessage("Success", "Created " + getDisplayEntityTypeName() + " " + getCurrentEntityInstanceName() + ".");
            resetListDataModel();
            current = newEntity;
            return prepareList();
        }
        catch (DmPortalException | RuntimeException ex) {
            SessionUtility.addErrorMessage("Error", "Could not create " + getDisplayEntityTypeName() + ": " + ex.getMessage());
            return null;
        }
    }

    public String prepareEdit(EntityType entity) {
        current = entity;
        return edit();
    }

    public String edit() {
        resetSelectDataModel();
        return "edit?faces-redirect=true";
    }

    protected void prepareEntityUpdate(EntityType entity) throws DmPortalException {
    }

    public String update() {
        try {
            logger.debug("Updating " + getDisplayEntityTypeName() + " " + getCurrentEntityInstanceName());
            prepareEntityUpdate(current);
            EntityType updatedEntity = getFacade().edit(current);
            SessionUtility.addInfoMessage("Success", "Updated " + getDisplayEntityTypeName() + " " + getCurrentEntityInstanceName() + ".");
            resetListDataModel();
            current = updatedEntity;
            return view();
        }
        catch (DmPortalException ex) {
            SessionUtility.addErrorMessage("Error", "Could not update " + getDisplayEntityTypeName() + ": " + ex.getMessage());
            return null;
        }
        catch (RuntimeException ex) {
            SessionUtility.addErrorMessage("Error", "Could not update " + getDisplayEntityTypeName() + ": " + getObjectAlreadyExistMessage(current));
            return null;
        }
    }
    
    abstract protected String getObjectAlreadyExistMessage(EntityType entity);

    public void destroy(EntityType entity) {
        current = entity;
        destroy();
    }

    public String destroy() {
        if (current == null) {
            logger.warn("Current item is not set");
            // Do nothing if current item is not set.
            return null;
        }
        try {
            logger.debug("Destroying " + getCurrentEntityInstanceName());
            getFacade().remove(current);
            SessionUtility.addInfoMessage("Success", "Deleted " + getDisplayEntityTypeName() + " " + getCurrentEntityInstanceName() + ".");
            resetListDataModel();
            return prepareList();
        }
        catch (Exception ex) {
            SessionUtility.addErrorMessage("Error", "Could not delete " + getDisplayEntityTypeName() + ": " + ex.getMessage());
            return null;
        }
    }

    public DataModel createListDataModel() {
        return new ListDataModel(getFacade().findAll());
    }

    public DataModel getListDataModel() {
        resetListDataModel();       
        listDataModel = createListDataModel();
        return listDataModel;
    }

    public void prepareEntityListForSelection(List<EntityType> selectEntityList) {
    }

    public DataModel createSelectDataModel() {
        List<EntityType> selectEntityList = getFacade().findAll();
        prepareEntityListForSelection(selectEntityList);
        return new ListDataModel(selectEntityList);
    }

    public DataModel getSelectDataModel() {
        if (selectDataModel == null) {
            selectDataModel = createSelectDataModel();
        }
        return selectDataModel;
    }

    public DataModel getItems() {
        return getListDataModel();
    }

    public List<EntityType> getSelectedObjectListAndResetSelectDataModel() {
        List<EntityType> returnList = selectedObjectList;
        resetSelectDataModel();
        return returnList;
    }

    public List<EntityType> getSelectedObjectList() {
        return selectedObjectList;
    }

    public List<EntityType> getFilteredObjectList() {
        return filteredObjectList;
    }

    public List<EntityType> getFilteredItems() {
        return filteredObjectList;
    }

    public void resetSelectedObjectList() {
        selectedObjectList = null;
    }

    public void setSelectedObjectList(List<EntityType> selectedObjectList) {
        this.selectedObjectList = selectedObjectList;
    }

    public void setFilteredObjectList(List<EntityType> filteredObjectList) {
        this.filteredObjectList = filteredObjectList;
    }

    public void setFilteredItems(List<EntityType> filteredItems) {
        this.filteredObjectList = filteredItems;
    }

    public void resetListDataModel() {
        listDataModel = null;
        listDataTable = null;
        filteredObjectList = null;
        current = null;
    }

    public void resetSelectDataModel() {
        selectDataModel = null;
        selectDataTable = null;
        selectedObjectList = null;
    }

    public List<EntityType> getAvailableItems() {
        return getFacade().findAll();
    }

    public EntityType getEntity(Integer id) {
        return getFacade().find(id);
    }

    public SelectItem[] getAvailableItemsForSelectMany() {
        return CollectionUtility.getSelectItems(getFacade().findAll(), false);
    }

    public SelectItem[] getAvailableItemsForSelectOne() {
        return CollectionUtility.getSelectItems(getFacade().findAll(), true);
    }

    public String getCurrentViewId() {
        return SessionUtility.getCurrentViewId();
    }

    public static String displayEntityList(List<?> entityList) {
        String itemDelimiter = ", ";
        return CollectionUtility.displayItemListWithoutOutsideDelimiters(entityList, itemDelimiter);
    }

}
